# ****************       BEGIN-STANDARD-COPYRIGHT      ***************
#
# Copyright (c) 2017, Spirent Communications.
#
# All rights reserved. Proprietary and confidential information of Spirent Communications.
#
#  ***************        END-STANDARD-COPYRIGHT       ***************

"""Session profile class and necessary tools."""

import six

from .internal.protocol import Type

from .session_response import SessionActionResponse
from .resources import ParameterFile, ResponseMapFile
from .identity import UriIdentity

_KNOWN_INVOKE_ARGS = ['response_map', 'parameters', 'command']

def _to_invoke_args(kwargs):
    """Prepares the given arguments dict for invoking as action arguments.

    Arguments:
    kwargs -- dict or arguments

    Returns a new dict, ready to be passed as arguments to invoke_command().
    """

    ret = {}
    params = {}

    for key, value in kwargs.items():
        if key in _KNOWN_INVOKE_ARGS:
            ret[key] = value
        else:
            params[key] = value

    if params:
        ret['parameters'] = params
    return ret

def _to_list(arg):
    """If arg is a list, returns it. If it's None, returns None. Otherwise, returns a one-element list with arg."""

    if arg is None:
        return None

    if isinstance(arg, list):
        return arg

    return [arg,]

def _to_requirements(agent_requirements):
    """Converts the given argument into a list of Type.Requirement.

    Arguments:
    agent_requirements -- dict of the format {'requirement_name': 'requirement_value'}. Can be None.

    Returns a list of Type.Requirement, or None.
    """

    if not agent_requirements:
        return None

    ret = []
    for key, value in agent_requirements.items():
        req = Type.Requirement()
        req.name = key
        req.value = value
        ret.append(req)

    return ret

def _to_property_group(properties):
    """Converts the given properties dict into Type.PropertiesGroup.

    Arguments:
        properties -- a dictionary with keys being property names (strings)
                      and value being property values (strings or nested properties group).
                      Can be None.

    Returns Type.PropertiesGroup, or None.
    """

    if not properties:
        return None

    ret = Type.PropertiesGroup()
    ret.name = ''

    for key, value in properties.items():
        if isinstance(value, six.string_types):
            # prop = Type.Property()
            prop = ret.properties.add()
            prop.name = key
            prop.value = value
        elif isinstance(value, list):
            for item in value:
                child = _to_property_group(item)
                child.name = key
                ret.children.extend([child])
        else:
            child = _to_property_group(value)
            if child:
                child.name = key
                ret.children.extend([child])

    return ret

def _to_params_list(parameters):
    """Converts the given parameters to a list of Type.Param.

    Arguments:
    parameters -- a dict with keys being parameter names and values being either parameter values, or
                  a dict with two keys: 'value', which is the parameter value, and 'children', which is a dict
                  of the same type as above. Example:
                  {'param1': 'value1', 'param2': {'value': 'value2', 'children': {'child_param1': 'value3'}}}
                  Can be None.

    Returns a list of Type.Param, or None.
    """

    if not parameters:
        return None

    ret = []

    for key, value in parameters.items():
        param = Type.Param()
        param.name = key
        if isinstance(value, ParameterFile):
            param.value = value.uri
        elif isinstance(value, ResponseMapFile):
            param.value = value.uri
        elif isinstance(value, dict):
            param.value = str(value['value'])
            if 'children' in value:
                param.parameters.extend(_to_params_list(value['children']))
        else:
            param.value = str(value)

        ret.append(param)

    return ret

class SessionProfileInformation(object):
    """ A details object to specify agent parameters"""

    def __init__(self, agent_name, agent_id, protocol_version, agent_type, capabilities ):
        """Initialise a agent parameters object
        agent_name -- a name of agent.
        agent_id -- a identifier of agent.
        protocol_version -- a version of protocol used
        """
        self.name = agent_id
        self.agent_name = agent_name
        self.protocol_version = protocol_version
        self.agent_type = agent_type
        self.capabilities = {}

        for cap in capabilities:
            cur_cap = self.capabilities.get(cap.name)
            if cur_cap is None:
                cur_cap = cap.value
            elif type(cur_cap) == list:
                cur_cap.append(cap.value)
            else:
                cur_cap = [cur_cap, cap.value]
            self.capabilities[cap.name] = cur_cap

    def __str__(self):
        return str(vars(self))

class SessionProfile(UriIdentity):
    """Session profile representation.

    Allows to open and close session, as well as invoking its actions and quickcalls.

    Recommended way to use this class is as follows:

        with proj.session_name_ffsp.open() as s1:
            # invoke a quickcall:
            response = s1.init_routes(all=True)

            # invoke a command:
            response = s1.command('command', response_map=proj.response_map_ffrm)

    Constructing instances of this class directly may be handy for testing, but generally is discouraged.

    Properties:
    agent -- agent information
    session_id -- if the session is opened, this will be a string representation of the session's ID.
                  Otherwise, this will be None. May be used to check is the session is opened.
    """

    def __init__(self, protocol_socket, agent_type, uri, dependencies=None):
        """Initiates a new session profile.

        Arguments:
        protocol_socket -- an instance of ProtocolSocket. It must be connected by the time open() is called.
        uri -- session profile URI, e.g. project://my_project/session_profiles/slc_test.ffsp
        dependencies -- session profile dependencies,
                        e.g. ['file:///home/dsavenko/itest/itest/dev/src/non-plugins/SpirentSLC/my_project.itar']

        """

        UriIdentity.__init__(self, uri)
        self._protosock = protocol_socket
        self._deps = dependencies

        self.agent = SessionProfileInformation(self._protosock.agent_name, self._protosock.agent_id, self._protosock.protocol_version, agent_type, self._protosock.capabilities)
        self.session_id = None
        self._quickcalls = None
        self.open_response = None

    def open(self, parameter_file=None, agent_requirements=None, properties=None, response_map_lib=None, **kwargs):
        """Opens the session.

        Arguments:
        parameter_file -- a single parameter file (URI), or a list of parameter files.
        agent_requirements -- dict of the format {'requirement_name': 'requirement_value'}
        properties -- a dictionary with keys being property names (strings)
                      and value being property values (strings or nested properties group).

        Raises ValueError, if the session is already opened. May raise socket.error.
        """

        if self.session_id:
            raise ValueError('Session already opened with id=' + self.session_id)

        if properties is None:
            properties = {}
            for key, value in kwargs.items():
                properties[key] = str(value)

        # Convert parameter file to uri
        if isinstance(parameter_file, ParameterFile):
            parameter_file = parameter_file.uri

        # Convert project to uri
        if isinstance(response_map_lib, UriIdentity):
            response_map_lib = response_map_lib.uri

        if response_map_lib != None and not (type(response_map_lib) in six.string_types):
            raise ValueError('Wrong response_map_lib value is passed. Should be URI' + str(response_map_lib))


        rsp = self._protosock.start_session(self._uri,
                                            dependencies=self._deps,
                                            param_files=_to_list(parameter_file),
                                            requirements=_to_requirements(agent_requirements),
                                            property_group=_to_property_group(properties),
                                            resp_map_lib=response_map_lib)
        self.session_id = rsp.sessionId
        self.open_response = SessionActionResponse(rsp)
        return self

    def close(self):
        """Closes the session.

        Does nothing, if the session has not been opened yet. May raise socket.error.
        """

        session_id = self.session_id
        self.session_id = None
        self.open_response = None
        if session_id:
            self._protosock.close_session(session_id)

    def is_open(self):
        """ Return if session is stil opened or not"""
        return self.session_id != None and self._protosock != None and self._protosock.is_open()

    def slc_query_properties(self):
        """Query a running session for all session properties"""
        return self.invoke_action('slc_query_properties')

    def invoke_action(self, action, command=None, parameters=None, response_map=None):
        """Invokes session action or quickcall.

        Generally, you do not need to use this method directly. E.g. instead of calling

            ssh_session.invoke_action('command', 'ls')

        you can just call

            ssh_session.command('ls')

        For Quickcalls, instead of calling

            session.invoke_action('my_quickcall', parameters={'param1': 'value1', 'param2': 123})

        you can just call

            session.my_quickcall(param1='value1', param2=123)

        Arguments:
        action -- session action or QuickCall name. Depends on a session.
                  Widely used actions: 'close' to close the session, 'command' to execute a specific session command.
        command -- command to execute
        parameters -- a dict with keys being parameter names and values being either parameter values, or
                      a dict with two keys: 'value', which is the parameter value, and 'children', which is a dict
                      of the same type as above. Example:
                      {'param1': 'value1', 'param2': {'value': 'value2', 'children': {'child_param1': 'value3'}}}
        responseMap -- URI of the response map file to use.

        Returns action response. Console calling is equivalent to response._text.
        Raises ValueError, if the session is not opened. May raise socket.error
        """

        if not self.session_id:
            raise ValueError('Session is not opened')

        # Convert response map to it uri
        if isinstance(response_map, ResponseMapFile):
            response_map = response_map.uri

        resp = self._protosock.invoke_action(self.session_id,
                                             action=action,
                                             command=str(command),
                                             params=_to_params_list(parameters),
                                             response_map=response_map)
        #if (isinstance(resp, InvokeResponse)):
        #    return resp.reportResult, resp.error
        #else:
        #    return SessionActionResponse(resp)
        return SessionActionResponse(resp)

    def list(self, qc_name=None):
        """Returns all QuickCalls available on a given session profile"""

        if qc_name:
            resp = self._protosock.query_session(self._uri, qc_name)
            qc = resp.quickCall[0]
            return dict((arg.name, arg.description) for arg in qc.args)

        if not self._quickcalls:
            self._quickcalls = dict()
            resp = self._protosock.query_session(self._uri)
            for qc in resp.quickCall:
                self._quickcalls[qc.name] = dict((arg.name, arg.description) for arg in qc.args)

        return self._quickcalls

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def __dir__(self):
        """ return a list of methods available to project"""
        res = super.__dir__(self) + list(self.list().keys())
        return res
    def __getitem__(self, key):
        return lambda *args, **kwargs: self.invoke_action(key, *args, **_to_invoke_args(kwargs))

    def __getattr__(self, key):
        ret = self[key]
        if not ret:
            raise AttributeError(key)
        return ret



# ****************       BEGIN-STANDARD-COPYRIGHT      ***************
#
# Copyright (c) 2017, Spirent Communications.
#
# All rights reserved. Proprietary and confidential information of Spirent Communications.
#
#  ***************        END-STANDARD-COPYRIGHT       ***************

from SpirentSLC import SLC

with SLC.init(host='localhost:9005', itar_path='itars/') as slc:
    proj = slc.open('NativePyLib')

    print('> 2. The Python user will start by connecting to one of the below iTest endpoints.  This will be known as an endpoint session.')
    print('>    - A running instance of the iTest GUI co-existing on the local machine or on a remote host over TCP/IP')
    print('>    - A temporal Velocity agent instance which is auto-launched by this library')
    print('> 4. Upon connect, the returned object would contain attributes about the iTest GUI or Agent instance.')
    print('-->')
    print(proj.dut_ffsp.agent)
    print('')

    print('> 5a. The Python user can specify a parameter file in a project or from the filesystem for the endpoint session (equivalent to setting a \'global parameter file\' in terms of setting parameter key priorities)')
    print('Without parameters:')
    print('-->')
    with proj.test_params_ffsp.open() as params_session:
        print(params_session.command('read 2'))

    print('With project://NativePyLib/test_parameters/parameters.ffpt')
    print('-->')
    with proj.test_params_ffsp.open(parameter_file='project://NativePyLib/test_parameters/parameters.ffpt') as params_session:
        print(params_session.command('read 2'))

    print('> 5c. The Python user can specify a local iTest topology file in a project to be used in the endpoint session')
    print('> 6. The Python user then attempts to open one or more test sessions (session profiles in projects or a topology devices\' sessions)')
    print('> 8. The Python user can issue native session-level commands using object oriented method syntax over the connection. These are built-in commands associated with open test sessions.')
    print('Command session \'help\' command:\n-->')
    cmd_session = proj.ServerAndPC_tbml.server1.cmd.open()
    print(cmd_session.command('help'))
    prompt_session = proj.dut_ffsp.open()
    print('Prompt  session \'help\' command:\n-->')
    print(prompt_session.command('help'))
    cmd_session.close()
    prompt_session.close()
    print('')

    print('> 9. The Python user can issue quickcalls associated with the opened test session using object oriented method syntax.')
    print('list of quickcalls:\n-->')
    print(proj.dut_ffsp.list())
    print('')

    print('Access the list of parameters for a specific QuickCall:\n-->')
    print(proj.dut_ffsp.list('init_routes'))
    print('')

    print('Topology session quickcalls:\n-->')
    with proj.ServerAndPC_tbml.server1.Process.open() as process_session:
        print(process_session.list())
        print(process_session.to_parent_dir())
    print('')



from SpirentSLC import SLC

"""Open the native ssh session type directly, supplying required session profile information"""

with SLC.init(host='localhost:9005') as slc:
    # # Process session
    # with slc.sessions.process.open() as process:
    #     print(process.command('help'))

    # # Command Prompt session
    # with slc.sessions.cmd.open() as cmd:
    #     print(cmd.command('dir'))

    
    # # HTTP session
    with slc.sessions.http.open() as http:
        resp = http.command('get page http://spirent.com')

        print('show queries')
        # Use automatic queties or response map queries
        print(resp.queries())
        # Call for particular query
        print(resp.pragma())

    # Rest session test
    # with slc.sessions.rest.open(url='https://jsonplaceholder.typicode.com/') as s1:
    #     resp = s1.GET('/posts/1')
    #     print(resp.json)

    # SSH session using properties tree
    ssh = slc.sessions.ssh.open(properties = {
       'ipAddress':'velocity-itest3.xored.com',
       'user':'root',
       'password':'iRyLc4KQj80=',
       'TerminalProperties':{
               'prompts': [ { 'prompt1':{ 'Content':'root@velocity-itest3:~#' } }, { 'prompt2':{ 'Content':'root@velocity-itest3:/#' } } ]
           }
       })
    print(ssh.command('cd /'))


# ****************       BEGIN-STANDARD-COPYRIGHT      ***************
#
# Copyright (c) 2017, Spirent Communications.
#
# All rights reserved. Proprietary and confidential information of Spirent Communications.
#
#  ***************        END-STANDARD-COPYRIGHT       ***************

"""Protocol implementation.

ProtocolSocket class provides low-level implementation of the Protobuf-based protocol communicating with agents.
"""

import uuid
import time
import socket
import six
import os

# pylint: disable=no-name-in-module,import-error
from google.protobuf.internal import encoder

from . import ExecutionProtocol_pb2 as pb

if six.PY3:
    # pylint: disable=redefined-builtin,invalid-name
    long = int

_PROTOCOL_LOGGING = 'True' == os.getenv('SLC_DEBUG_PROTOCOL', 'False')

_EXECUTION_MESSAGE_SUBTYPE_INFO = {
    # pylint: disable=bad-whitespace

    # This map is used to wrap messages into ExecutionMessage. The keys are messages classes, and values are
    # pairs (message type, message field name) in ExecutionMessage.
    # It's safer not to use it directly, but use _get_msg_subtype_info() instead.

    # EXECUTION MESSAGES:

    pb.CancelMessage:           (pb.ExecutionMessage.CANCEL_MESSAGE,            'cancelMessage'),
    pb.HeartbeatMessage:        (pb.ExecutionMessage.HEARTBEAT_MESSAGE,         'heartbeatMessage'),
    pb.InitMessage:             (pb.ExecutionMessage.INIT_MESSAGE,              'initMessage'),
    pb.InvokeAck:               (pb.ExecutionMessage.INVOKE_ACK,                'invokeAck'),
    pb.InvokeMessage:           (pb.ExecutionMessage.INVOKE_MESSAGE,            'invokeMessage'),
    pb.InvokeResponse:          (pb.ExecutionMessage.INVOKE_RESPONSE,           'invokeResponse'),
    pb.ReportMessage:           (pb.ExecutionMessage.REPORT_MESSAGE,            'reportMessage'),
    pb.StepMessage:             (pb.ExecutionMessage.STEP_MESSAGE,              'stepMessage'),
    pb.TerminateMessage:        (pb.ExecutionMessage.TERMINATE_MESSAGE,         'terminateMessage'),
    pb.TokenMessage:            (pb.ExecutionMessage.TOKEN_MESSAGE,             'tokenMessage'),
    pb.ExecutionIssueMessage:   (pb.ExecutionMessage.EXECUTION_ISSUE_MESSAGE,   'executionIssueMessage'),

    # SESSION CONTROL MESSAGES:

    pb.NewSession:                  (pb.ExecutionMessage.NEW_SESSION_MESSAGE,           'newSessionMessage'),
    pb.NewSessionResponse:          (pb.ExecutionMessage.NEW_SESSION_RESPONSE,          'newSessionResponse'),
    pb.SendTerminalData:            (pb.ExecutionMessage.SEND_TERMINAL_DATA_MESSAGE,    'sendTerminalDataMessage'),
    pb.SendTerminalDataResponse:    (pb.ExecutionMessage.SEND_TERMINAL_DATA_RESPONSE,   'sendTerminalDataResponse'),
    pb.SessionAction:               (pb.ExecutionMessage.SESSION_ACTION_MESSAGE,        'sessionActionMessage'),
    pb.SessionActionResponse:       (pb.ExecutionMessage.SESSION_ACTION_RESPONSE,       'sessionActionResponse'),
    pb.ListProjects:                (pb.ExecutionMessage.LIST_PROJECTS,                 'listProjects'),
    pb.ListProjectsResponse:        (pb.ExecutionMessage.LIST_PROJECTS_RESPONSE,        'listProjectsResponse'),
    pb.QueryProject:                (pb.ExecutionMessage.QUERY_PROJECT,                 'queryProject'),
    pb.QueryProjectResponse:        (pb.ExecutionMessage.QUERY_PROJECT_RESPONSE,        'queryProjectResponse'),
    pb.QuerySession:                (pb.ExecutionMessage.QUERY_SESSION,                 'querySession'),
    pb.QuerySessionResponse:        (pb.ExecutionMessage.QUERY_SESSION_RESPONSE,        'querySessionResponse')
}

class ProtocolError(socket.error):
    """Any protocol-related error."""
    pass

def _get_msg_subtype_info(klass):
    if klass not in _EXECUTION_MESSAGE_SUBTYPE_INFO:
        raise ValueError('Unrecognized message class ' + str(klass))

    return _EXECUTION_MESSAGE_SUBTYPE_INFO[klass]

def _recv_bytes(sock, length):
    """Receives the given number of bytes from socket.

    Arguments:
    sock -- socket to receive from.
    length -- number of bytes to receive.

    Returns the bytes as a byte string. Blocks until the given number of bytes is read. Raises socket.error, if
    the socket is closed prematurely.
    """

    chunks = []
    bytes_recd = 0
    while bytes_recd < length:
        chunk = sock.recv(min(length - bytes_recd, 8192))
        if chunk == b'':
            raise socket.error("Socket connection broken")
        chunks.append(chunk)
        bytes_recd = bytes_recd + len(chunk)
    return b''.join(chunks)

def _send_bytes(sock, msg):
    """Sends the given message (string) to the socket.

    Arguments:
    sock -- socket to send to.
    msg -- string to send.

    Blocks until the message is sent fully. Raises socket.error, if the socket is closed prematurely.
    """

    length = len(msg)
    total = 0
    while total < length:
        sent = sock.send(msg[total:])
        if sent == 0:
            raise socket.error("Socket connection broken")
        total = total + sent

def _read_varint(sock):
    """Reads varint from socket.

    Originally taken from google.protobuf.internal.decoder._VarintDecoder and modified for reading varints from sockets.

    Arguments:
    sock -- socket to read from.

    Returns pair (value, bytes_read). Raises ValueError on invalid varint.
    """

    mask = (1 << 64) - 1
    result_type = long

    result = 0
    shift = 0
    pos = 0
    # pylint: disable=superfluous-parens,invalid-name,bad-option-value,redefined-variable-type
    while 1:
        b = six.indexbytes(_recv_bytes(sock, 1), 0)
        result |= ((b & 0x7f) << shift)
        pos += 1
        if not (b & 0x80):
            result &= mask
            result = result_type(result)
            return (result, pos)
        shift += 7
        if shift >= 64:
            raise ValueError('Too many bytes when decoding varint.')

def _ensure_msg_type(msg, expected_type):
    if msg.type != expected_type:
        raise ProtocolError('Expected to receive ' + msg.Type.Name(expected_type) + ' from agent, but received ' +
                            msg.Type.Name(msg.type))

def _unwrap(msg):
    for k in _EXECUTION_MESSAGE_SUBTYPE_INFO:
        msg_type, msg_field = _EXECUTION_MESSAGE_SUBTYPE_INFO[k]
        if msg_type == msg.type:
            return getattr(msg, msg_field)
    raise ValueError('Unrecognized message type ' + str(msg.type))

def _read_delimited_msg(sock):
    """Reads delimited message from socket.

    Arguments:
    sock -- socket to read from.
    unwrapToKlass -- if not None, tries to unwrap the received message to the given class. See unwrap() for details.
                     Otherwise, return ExecutionMessage.
                     Default: None.

    Returns the message.
    """

    data_len, _ = _read_varint(sock)
    data = _recv_bytes(sock, data_len)
    ret = pb.ExecutionMessage()
    ret.ParseFromString(data)
    return ret

def _wrap(msg):
    """Wraps the given message into ExecutionMessage.

    Arguments:
    msg -- message to wrap.

    Returns ExecutionMessage, which wrapped the given message. If the given message is already an ExecutionMessage,
    just returns the given message.

    Raises ValueError, if the given message is not of a recognized message type.
    """

    if isinstance(msg, pb.ExecutionMessage):
        return msg

    msg_type, msg_field = _get_msg_subtype_info(type(msg))

    ret = pb.ExecutionMessage()
    ret.type = msg_type
    getattr(ret, msg_field).CopyFrom(msg)

    return ret

def _write_delimited_msg(sock, msg):
    """Writes delimited message to the socket.

    Arguments:
    sock -- socket to write to.
    msg -- protobuf message to write. If it's not already wrapped, it will be wrapped first. See wrap().

    May raise socket.error.
    """

    wrapped = _wrap(msg)

    if _PROTOCOL_LOGGING:
        print("_write:", wrapped)

    msg_str = wrapped.SerializeToString()

    # pylint: disable=protected-access
    delim = encoder._VarintBytes(len(msg_str))

    _send_bytes(sock, delim + msg_str)

def _force_connect(host, port, timeout, agent_process):
    """Constantly tries to connect to the given host/port pair for the given timeout in seconds (approximately).

    Arguments:
    host -- hostname or IP
    port -- port number
    timeout -- timeout in seconds

    If connection is successful, returns the connected socket. If the timeout is passed, but connection is not
    established, raises a TimeoutError.
    """

    start_time = time.time()
    attempt_counter = 0
    while time.time() - start_time < timeout:
        try:
            attempt_counter += 1
            return socket.create_connection((host, port))
        except IOError:
            time.sleep(0.1)

        if agent_process != None:
            pp = agent_process.poll()
            if pp != None:
                raise Exception('Velocity agent process is terminated')

        if attempt_counter == 30:
            print('trying to connect to host', host)
            attempt_counter = 0
    raise TimeoutError('Fail to connect to host %(host)s port:%(port)d in %(timeout)d sec.' %{"host": host, "timeout": timeout, "port": port})


class Type(object):
    """Directly expose some PB classes to the user. It's really the most reasonable way."""

    # pylint: disable=invalid-name
    Param = pb.Param
    Requirement = pb.NewSession.Requirement
    PropertiesGroup = pb.PropertiesGroup
    Property = pb.PropertiesGroup.Property

class ProtocolSocket(object):
    """This class implements Protobuf-based protocol of communication with agents.

    Unless specified otherwise, this class's methods, including the constructor, may raise socket.error and
    ProtocolError (which is a sub-class of socket.error).
    """

    def __init__(self, host, port, timeout=60, agent_process = None):
        """Creates a new protocol socket.

        Arguments:
        host -- hostname (or IP address) to connect to.
        port -- port number.
        timeout -- timeout in seconds.

        It tries to connect to the given host/port pair for the given timeout in seconds (approximately).
        If connection is successful, a new ProtocolSocket is returned. Otherwise, socket.error is raised.
        If connection is successful, but the peer doesn't behave as expected (probably, it's not an agent),
        ProtocolError is raised.

        Do not forget to call close(), when you're done.
        """
        self._sock = None
        self.agent_name = None
        self.agent_id = None
        self.protocol_version = None

        try:
            self._sock = _force_connect(host, port, timeout, agent_process)
            msg = _read_delimited_msg(self._sock)
            _ensure_msg_type(msg, msg.INIT_MESSAGE)
            self.agent_id = msg.initMessage.agentId
            self.agent_name = msg.initMessage.agentName
            self.protocol_version = msg.initMessage.protocolVersion

            #  Store capability
            self.capabilities = msg.initMessage.capability
        except socket.error as err:
            self.close()
            raise

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def is_open(self):
        """ Return true if connection is alive """
        return self._sock != None

    def close(self):
        """Closes the protocol socket."""

        sock = self._sock
        self._sock = None
        if sock:
            try:
                # sock.shutdown(socket.SHUT_RDWR)
                sock.close()
            except socket.error as err:
                print("failed with socket close", err )

    def _read(self, predicate):
        """Reads messages until the given predicate is True.

        Arguments:
        predicate -- a function with one argument (ExecutionMessage) and a boolean return value. If it returns True,
                     this method stops and returns this message. If it returns False, this method reads the next
                     message.

        Returns the messages read on which the predicate is True.
        """

        while True:
            msg = _read_delimited_msg(self._sock)
            if predicate(msg):
                if _PROTOCOL_LOGGING:
                    print("_read:", msg)
                msg = _unwrap(msg)
                return msg

    def _write(self, msg):
        """Writes the given message to socket.

        Arguments:
        msg -- message to write. If it's not an instance of ExecutionMessage, it will be wrapped to ExecutionMessage.
        """
        _write_delimited_msg(self._sock, msg)

    def start_session(self, session_uri, dependencies=None, param_files=None, requirements=None, property_group=None, resp_map_lib=None):
        """Starts a new session.

        Arguments:
        session_uri -- session URI, e.g. 'project://my_project/session_profiles/ssh_session.ffsp'.
        dependencies -- a list of project URIs this session depend on, e.g. ['file:///tmp/ws/my_project.itar'].
        param_files -- a list of parameter file URIs.
        requirements -- a list of Type.Requirement instances.
        property_group -- a Type.PropertyGroup instance.
        resp_map_lib -- URI of Response Map Library. i.e 'project://map_library/'

        Returns a new session response.
        """

        try:
            msg = pb.NewSession()
            msg.sessionId = str(uuid.uuid1())
            msg.sessionUri = session_uri
            if dependencies:
                msg.dependencyUri.extend(dependencies)
            if param_files:
                msg.paramFileUri.extend(param_files)
            if requirements:
                msg.requirement.extend(requirements)
            if property_group:
                msg.properties.CopyFrom(property_group)
            if resp_map_lib:
                msg.responseMapLib = resp_map_lib
            self._write(msg)
            ret = self._read(lambda msg: msg.type == msg.NEW_SESSION_RESPONSE)
            ret.sessionId = msg.sessionId
            return ret
        except socket.error as err:
            self.close()
            raise

    def invoke_action(self, session_id, action, command=None, params=None, response_map=None):
        """Executes a session action or QuickCall and returns the result.

        Arguments:
        session_id -- session ID (string). The session must be started first. See 'start_session()'.
        action -- session action or QuickCall name. Depends on a session.
                  Widely used actions: 'close' to close the session, 'command' to execute a specific session command.
        command -- command to execute
        params -- a list of Type.Param instances.
        responseMap -- URI of the response map file to use.

        Returns session action response.
        """

        if self._sock is None:
            raise IOError("Connection socket closed")

        try:
            msg = pb.SessionAction()
            msg.sessionId = session_id
            msg.action = action
            if command:
                msg.command = command
            if params:
                msg.parameters.extend(params)
            if response_map:
                msg.responseMapMode = msg.FILE
                msg.responseMap = response_map
            self._write(msg)
            return self._read(lambda msg: msg.type in [msg.SESSION_ACTION_RESPONSE, msg.INVOKE_RESPONSE])
        except:
            self.close()
            raise

    def close_session(self, session_id):
        """Closes session with the given ID.

        Arguments:
        session_id -- session ID (string). The session must be started first. See 'start_session()'.

        Returns session action response.
        """

        return self.invoke_action(session_id, 'close')

    def invoke_command(self, session_id, command):
        """Executes a session command.

        Arguments:
        session_id -- session ID (string). The session must be started first. See 'start_session()'.
        command -- command to execute.

        Returns session action response.
        """

        return self.invoke_action(session_id, 'command', command=command)

    def list_projects(self):
        """Returns a list of project names that are accessible to the current user (if authenticated through Velocity) within the applicable context.
             - For a local agent it should list all projects that are available within the ITAR_PATH.
             - For a iTest GUI instance it should list all projects in the workspace.
             - For Velocity, it should list all projects available within all repos that the user has access to.

             When working through Velocity the names include the repo path ('/main/') within them.
             When working outside of Velocity only the project name will be returned. I.e. 'my_project'.
        """
        try:
            msg = pb.ListProjects()
            self._write(msg)
            ret = self._read(lambda msg: msg.type == msg.LIST_PROJECTS_RESPONSE)

            return ret.projects
        except:
            self.close()
            raise

    def query_project(self, project, session_profile=True, topology=True, parameter_file=False, response_map=False):
        """Returns all usable resources in project: session profiles, topologies, property files and response maps.

        Arguments:
        project -- name of project (string).
        session_profile -- Include the list of session profiles to result (boolean). Default True.
        topology -- Include the list of topologies to result (boolean). Default True.
        parameter_file -- Include the list of parameter files to result (boolean). Default False.
        response_map -- Include the list of response maps to result (boolean). Default False.
        """

        try:
            msg = pb.QueryProject()
            msg.project = project
            msg.sessionProfiles = session_profile
            msg.topologies = topology
            msg.parameterFiles = parameter_file
            msg.responseMaps = response_map
            self._write(msg)

            return self._read(lambda msg: msg.type == msg.QUERY_PROJECT_RESPONSE)
        except:
            self.close()
            raise

    def query_session(self, session_uri, qc_name=None):
        """Returns all QuickCalls available on a given session

        Arguments:
        session_uri -- session URI (string).
        """

        try:
            msg = pb.QuerySession()
            msg.sessionUri = session_uri

            if qc_name:
                msg.specificQC = qc_name
            self._write(msg)

            return self._read(lambda msg: msg.type == msg.QUERY_SESSION_RESPONSE)
        except:
            self.close()
            raise



# ****************       BEGIN-STANDARD-COPYRIGHT      ***************
#
# Copyright (c) 2017, Spirent Communications.
#
# All rights reserved. Proprietary and confidential information of Spirent Communications.
#
#  ***************        END-STANDARD-COPYRIGHT       ***************

""" Private SLC configuration access functions."""

import os

def _get_int_var(name, default):
    """Returns the value an environment variable with the given name as integer, or the default value.

    Arguments:
    name -- env var name
    default -- default value

    Returns the value of the varibale, if it exists and can be converted to integer. Otherwise, returns the default
    value.
    """

    try:
        return int(os.getenv(name, default))
    except ValueError:
        return default

def get_agent_path():
    """Returns the agent location as a path to the Velocity agent folder."""
    return os.getenv('SPIRENT_SLC_AGENT_PATH', 'velocity-agent')

def get_agent_host_port():
    """Returns 'host:port' string, which points to the SLC host.

    May return 'local', which means to run the standalone host.
    """
    return os.getenv('SPIRENT_SLC_HOST', 'local')

def get_local_agent_host_port():
    """Returns 'host:port' string, which specifies on which network interface/port the standalone agent should bind."""
    return os.getenv('SPIRENT_SLC_LOCAL_HOST', 'localhost:9002')

def get_agent_password():
    """Returns the password to connect to an SLC host.

    May return None, in which case no password is required.
    """
    return os.getenv('SPIRENT_SLC_PASSWORD')

def get_agent_start_timeout():
    """Returns for how many seconds we should wait for the standalone agent to start before concluding the agent is
    failed to start.
    """
    return _get_int_var('SPIRENT_SLC_AGENT_START_TIMEOUT', 60)

def get_itar_path():
    """Returns the path to folder where iTars are placed."""
    res = os.getenv('ITAR_PATH')
    if res is None:
        return os.getcwd()
    return res

def get_license_server():
    """Returns license server in a form "host:port"""
    server = os.getenv('SPIRENT_SLC_LICENSE_SERVER')
    if server != None and len(server.strip()) == 0:
        return None
    return server

def get_license_proxy():
    """Returns license proxy in a form "host:port"""
    proxy = os.getenv('SPIRENT_SLC_LICENSE_PROXY')
    if proxy != None and len(proxy.strip()) == 0:
        return None
    return proxy

def is_verbose_mode():
    """Returns True if SPIRENT_SLC_VERBOSE proprty is defined """
    return os.getenv('SPIRENT_SLC_VERBOSE') != None



# ****************       BEGIN-STANDARD-COPYRIGHT      ***************
#
# Copyright (c) 2017, Spirent Communications.
#
# All rights reserved. Proprietary and confidential information of Spirent Communications.
#
#  ***************        END-STANDARD-COPYRIGHT       ***************
"""
A setup script for SpirentSLC
"""
import codecs
from os import path

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding

# Get the long description from the README file
def _get_long_description():
    here = path.abspath(path.dirname(__file__))
    with codecs.open(path.join(here, 'README.md'), encoding='utf-8') as readme_file:
        return readme_file.read()
    return ""

setup(
    name='SpirentSLC',
    version='1.0.0.dev14',
    author='Spirent.com',
    author_email='itest@spirent.com',

    description='Spirent Session Control Library',
    long_description=_get_long_description(),

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',

        'Topic :: Software Development',
        'Topic :: Software Development :: Testing',

         # Pick your license as you wish (should match "license" above)
        'License :: Other/Proprietary License',

        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],

    packages=find_packages(exclude=['contrib', 'docs', 'tests']),

    install_requires=['protobuf'],

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    # extras_require={
    #     'dev': ['check-manifest'],
    #     'test': ['coverage'],
    # },

    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    package_data={
        'itars': ['itars/*.itar', 'README.md'],
    },
)



# ****************       BEGIN-STANDARD-COPYRIGHT      ***************
#
# Copyright (c) 2017, Spirent Communications.
#
# All rights reserved. Proprietary and confidential information of Spirent Communications.
#
#  ***************        END-STANDARD-COPYRIGHT       ***************

"""Parameter file class and necessary tools."""

from .identity import UriIdentity

class ParameterFile(UriIdentity):
    """Parameter file representation.
    """
    def __init__(self, uri):
        """Initiates a new topology.

        Arguments:
        uri -- parameter file URI, e.g. project://my_project/parameter_files/parameters.ffpt
        """

        UriIdentity.__init__(self, uri)

    def __str__(self):
        """ Return a URI """
        return self._uri

class ResponseMapFile(UriIdentity):
    """Responsemap file representation.
    """
    def __init__(self, uri):
        """Initiates a new topology.

        Arguments:
        uri -- parameter file URI, e.g. project://my_project/response_maps/my_map.ffrm
        """
        UriIdentity.__init__(self, uri)

    def __str__(self):
        """ Return a URI """
        return self._uri
